/**
 * This is the Message class.
 */

package salsa.language;

import java.lang.reflect.Method;

import java.net.InetAddress;
import java.net.UnknownHostException;

import wwc.naming.UAL;
import wwc.messaging.Theater;

public class Message implements java.io.Serializable {
	/**
		Actor that the message is being sent from.
	*/
	private Actor sender;

	/**
		Actor that the message is being sent to.
	*/
	private Actor target;

	/**
		Name of the method that should be invoked.
	*/
	private String methodName;

	/**
		Array of values that are copied to be the arguments of the method. If one of
		the arguments is the value of a yet to be evaluated token an instance of the
		Token class should be subituted for it until it is evaluated.
	*/
	private Object[] arguments;

	/**
		Reference to the next message in the continuation.
	*/
	private Message continuation;

	/**
		This is a simple way to output the name of a message in a
		readable format. used for testing purposes.
	*/
	public String getName() {
		String text = "[Source: ";
		if (sender != null) text += sender.getName();
		else text += "null";

		text += ", Target: ";
		if (target != null) text += target.getName();
		else text += "null";

		text += ", Method: " + methodName;

		if (arguments != null) {
			text += ", Arguments: {";
			for (int i = 0; i < arguments.length; i++) {
				if (arguments[i] != null) text += arguments[i].getClass().getName();
				else text += "null";

				if (i != (arguments.length-1)) text += ", ";
			}
			text += "}";
		}
		else text += null;

//		if (continuation != null) text +=  ", Continuation: " + continuation.getName();

		if (continuation != null) text += ", Continuation: true";
		else text += ", Continuation: false";

		text += "]";
		return text;
	}


	/**
		This constructor takes a sender, and three arrays containing respective targets,
		message names and argument arrays.
		The constructor creates a linked list of messages for each message in the continuation.
	*/
	public Message( Actor sender, Actor[] targets, String[] methodNames, Object[][] arguments) {

		this( sender, targets[ 0 ], methodNames[ 0 ], arguments[ 0 ] );

		Message m = this;

		for (int i=1;i<targets.length;++i) {
			m.continuation = new Message( targets[ i-1 ], targets[ i ], methodNames[ i ], arguments[ i ] );
			m = m.continuation;
		}
	}

	/**
		The constructor takes a sender, target, message name, and array of arguments and
		constructs one Message object. The continuation field remains null.
	*/
	public Message( Actor sender, Actor target, String methodName, Object[] arguments) {
		this.sender = sender;
		this.target = target;
		this.methodName = methodName;
		this.arguments = arguments;
	}

	/**
		This method sets any arguments which are instances of Token to the specified reference.
	*/
	public void updateTokens( Object newToken ) {
		for (int i=0; i<arguments.length; ++i) {
			if (arguments[i] instanceof Token) arguments[i] = newToken;
		}
	}

   /**
      This method returns the target of this message.
   */
   public Actor getTarget() {

      return target;
   }

   /**
      This method returns the sender of this message.
   */
   public Actor getSource() {

      return sender;
   }

   /**
      This method returns the arguments of the message.
   */
   public Object[] getArguments() {

      return arguments;
   }

   /**
      This method returns the continuation of the message or null
      if none exists.
   */
   public Message getContinuation() {

      return continuation;
   }

	/**
		This recursive method is called by the MessageGroup class that
		implements the Join Director. It adds the specified message to the
		end of the continuation list.
	*/
	protected void addContinuation( Message newContinuation ) {

		if (continuation == null) {
			continuation = newContinuation;

		} else {
			continuation.addContinuation( newContinuation );
		}
	}


   /**
      This method returns the name of the message.
   */
   public String getMethodName() {

      return methodName;
   }

	/**
		This method checks to see if the source and target are at the current theater and if so
		updates their references to point to the local copies.
	*/
	public void resolveActors(Theater theater) {

      // Check to see if target and sender are local.
      // If so simply reference the already existing actor.
      UAL senderUal = ((UniversalActor)target).getUAL();
      UAL targetUal = ((UniversalActor)target).getUAL();

      String localHost = null;
      try {
      	localHost = InetAddress.getLocalHost().getHostName();

     	} catch (UnknownHostException e) {
     		localHost = null;
     	}

      if (targetUal != null &&
         targetUal.getHost().equals(localHost) &&
         targetUal.getPort() == theater.getPort() &&
         theater.getActorUAL(targetUal.getIdentifier()) != null) {

         target = theater.getActorUAL(targetUal.getIdentifier());
      }

      if (senderUal != null &&
         senderUal.getHost().equals(localHost) &&
         senderUal.getPort() == theater.getPort() &&
         theater.getActorUAL(senderUal.getIdentifier()) != null) {

         sender = theater.getActorUAL(senderUal.getIdentifier());
      }
	}
}




